from dataclasses import dataclass
from typing import List, Optional, Union, Dict, Tuple
from tqdm import tqdm
from .utils import *
from .info_extractor import SquadAugmentedExample, InfoConfig, extract_clue_and_question_info, extract_clue, extract_question_type_and_id, NLP
from collections import Counter
import os
import multiprocess as mp
import math
import numpy as np
import pickle as pkl

@dataclass
class SquadExample:
    """
    A single example for the Squad-Zhou dataset, as loaded from disk.
    """
    context_text: str
    question_text: str
    answer_text: str
    answer_start: int


@dataclass
class SquadProcessedExample:
    """
    A single example for the processed SQuAD Zhou dataset, used for training.
    """
    context_text: str
    question_text: str
    question_type: str
    answer_text: str
    answer_start: int
    clue_text: str
    clue_start: int
    para_id: int


def read_squad_examples(input_file: str) -> List[SquadExample]:
    """
    Read a SQuAD Zhou text file into a list of SquadExample.
    """
    raw_examples = []
    with open(input_file, encoding="utf-8") as fh:
        lines = fh.readlines()
        for line in tqdm(lines):
            fields = line.strip().split("\t")
            input_sent, answer_text, question_text = fields[6], fields[8], fields[9]
            answer_start_token = int(fields[1].split()[0])

            # Calculate token spans and answer start in the tokenized sentence
            token_spans = find_token_spans_in_text(
                fields[0], fields[0].split())
            answer_start_in_tokenized_sent = token_spans[answer_start_token][0]

            # Find all answer spans in the input sentence
            answer_spans = match_spans(answer_text, input_sent)
            # skip if no answer spans are found, the answer does does not exist in the input sentence
            if len(answer_spans) == 0:
                continue
            # Find the closest answer span to the answer start in the tokenized sentence
            answer_start = min(
                answer_spans,
                key=lambda span: abs(span[0] - answer_start_in_tokenized_sent),
            )[0]

            example = SquadExample(
                question_text=normalize(question_text),
                context_text=normalize(input_sent),
                answer_text=normalize(answer_text),
                answer_start=answer_start,

            )
            raw_examples.append(example)

    return raw_examples


def create_squad_example_with_info(raw_ex: List[SquadExample]) -> List[SquadAugmentedExample]:
    """
    Augment the raw examples with question-type and clue info.
    """

    num_process = 1
    start_index = 0
    end_index = len(raw_ex)
    batch_size = len(raw_ex) // num_process


    def task(j):
        start = start_index + j * batch_size
        end = min(start_index + (j + 1) * batch_size, end_index)
        examples = []
        e: SquadExample
        for e in tqdm(raw_ex[start:end], desc=f"Process {j}", position=j, leave=False):
            new_e = extract_clue_and_question_info(
                sentence=e.context_text, question=e.question_text, answer=e.answer_text, answer_start=e.answer_start)
            examples.append(new_e)
        return examples
    
    # examples_list = []
    # with mp.Pool(num_process) as pool:
    #     examples_list = pool.map(task, range(num_process))
    
    examples_with_info = task(0)
    # for e in examples_list:
        # examples_with_info += e

    return examples_with_info


def calculate_probability_distribution(augmented_examples: List[SquadAugmentedExample]) -> Dict[str, Counter]:
    """
    Calculates the probability distribution of answer, clue, and sentence based on the given list of augmented examples.

    The probability distribution is defined as:
    P(a, c, s) = p(a) * p(c|a) * p(s|c, a)
               = p(a|a_tag, a_length) * p(c|c_tag, dep_dist) * p(s|a_tag)

    Args:
        augmented_examples (List[SquadAugmentedExample]): A list of SquadAugmentedExample objects.

    Returns:
        Dict[str, Counter]: A dictionary containing the probability distribution of answer, clue, and sentence.
    """

    """
    Disclaimer this function is adapted from the original implementation
    """
    sla_tag = []
    clc_tag_dep_dist = []
    ala_tag_a_length = []

    for e in tqdm(augmented_examples):
        a_tag = "-".join([e.answer_pos_tag, e.answer_ner_tag])
        s = e.question_type  # question style (type)
        a_length = e.answer_length
        a_length_bin = value_to_bin(
            a_length,
            InfoConfig.ans_len_min_val,
            InfoConfig.ans_len_max_val,
            InfoConfig.ans_len_bin_width,
        )
        c_tag = "-".join([e.clue_info.clue_pos_tag, e.clue_info.clue_ner_tag])
        dep_dist = e.clue_info.clue_answer_dep_path_len
        dep_dist_bin = value_to_bin(
            dep_dist,
            InfoConfig.clue_dep_dist_min_val,
            InfoConfig.clue_dep_dist_max_val,
            InfoConfig.clue_dep_dist_bin_width,
        )
        sla_tag.append("_".join([s, a_tag]))
        clc_tag_dep_dist.append("_".join([c_tag, str(dep_dist_bin)]))
        ala_tag_a_length.append("_".join([a_tag, str(a_length_bin)]))

    sla_tag = Counter(sla_tag)
    clc_tag_dep_dist = Counter(clc_tag_dep_dist)
    ala_tag_a_length = Counter(ala_tag_a_length)
    sample_probs = {
        "a": ala_tag_a_length,
        "c|a": clc_tag_dep_dist,
        "s|c,a": sla_tag,
    }

    return sample_probs


def create_process_squad_examples(raw_ex: List[SquadExample]):
    """
    Get a list of spaCy processed examples.
    """

    raw_ex = list(enumerate(raw_ex))
    start_index = 0
    end_index = len(raw_ex)
    batch_size = 10000
    num_batches = math.ceil((end_index - start_index) / batch_size)

    def task(j):
        start = start_index + j * batch_size
        end = min(start_index + (j + 1) * batch_size, end_index)
        examples = []
        e: SquadExample
        for pid, e in tqdm(raw_ex[start:end], desc=f"Process {j}", position=j, leave=False):

            context_spacydoc = NLP(e.context_text)
            context_tokens = [token.text for token in context_spacydoc]
            spans = find_token_spans_in_text(e.context_text, context_tokens)

            question_spacydoc = NLP(e.question_text)
            ques_type, ques_type_id = extract_question_type_and_id(
                e.question_text)

            answer_start = e.answer_start
            answer_end = answer_start + len(e.answer_text)

            answer_span = []
            for idx, span in enumerate(spans):
                if not (answer_end <= span[0] or answer_start >= span[1]):
                    answer_span.append(idx)

            y1_in_sent, y2_in_sent = answer_span[0], answer_span[-1]
            answer_in_sent = " ".join(
                context_tokens[y1_in_sent: y2_in_sent + 1])

            clue_info = extract_clue(
                e.context_text, e.question_text, answer_in_sent, y1_in_sent)

            sent_limit = InfoConfig.sent_limit
            clue_token_position = np.where(
                clue_info.padded_selected_clue_binary_ids == 1)[0]

            if len(clue_token_position) > 0 and clue_info.clue_chunk:
                start_idx = clue_token_position[0]
                end_idx = clue_token_position[-1]
                if len(spans) <= end_idx:
                    # num_spans_len_error += 1
                    clue_text, clue_start = None, None
                else:
                    start, end = spans[start_idx][0], spans[end_idx][1]
                    clue_text = e.context_text[start:end]
                    clue_tokenized_text = " ".join(
                        clue_info.clue_chunk[2])
                    if clue_text != clue_tokenized_text:
                        clue_start = e.context_text.find(clue_tokenized_text)
                        clue_text = clue_info.clue_chunk[2][0]
                        if clue_start < 0:
                            continue
                    else:
                        clue_start = start

            else:
                clue_text, clue_start = None, None

            example = SquadProcessedExample(
                context_text=e.context_text,
                question_text=e.question_text,
                answer_text=e.answer_text,
                answer_start=e.answer_start,
                question_type=ques_type,
                clue_text=clue_text,
                clue_start=clue_start,
                para_id= pid
            )
            examples.append(example)
        return examples

    acc = []
    # with mp.Pool(4) as pool:
        # for processed_examples in pool.imap(task, range(num_batches)):
            # acc += processed_examples
    acc = task(0)
    return acc


def pipeline(input_file: str):
    """
    Pipeline for processing squad examples.
    """
    mp.set_start_method("spawn")

    raw_ex = read_squad_examples(input_file)
    raw_ex = raw_ex[:1000]

    processed_ex = create_process_squad_examples(raw_ex)
    print(processed_ex[:6])
    with_info_ex = create_squad_example_with_info(raw_ex[:1000])

    sample_probs = calculate_probability_distribution(with_info_ex)

    return sample_probs


if __name__ == "__main__":
    data = pipeline("botiverse/Theorizer/squad/dataset/train.txt")
    print(data)


'''
Common constants
'''
import os

THEORIZER_PATH= os.path.abspath(os.path.dirname(__file__))
BOTIVERSE_PATH=None
SQUAD_TRAIN_PATH=None

QUESTION_TYPES = [
    "Who", "Where", "When", "Why", "Which", "What", "How", "Yes-No", "Other"]
INFO_QUESTION_TYPES = [
    "Who", "Where", "When", "Why", "Which", "What", "How"]
YES_NO_QUESTION_TYPES = [
    "Am", "Is", "Was", "Were", "Are",
    "Does", "Do", "Did",
    "Have", "Had", "Has",
    "Could", "Can",
    "Shall", "Should",
    "Will", "Would",
    "May", "Might"]
Q_TYPE2ID_DICT = {
    "What": 0, "Who": 1, "How": 2,
    "Where": 3, "When": 4, "Why": 5,
    "Which": 6, "Boolean": 7, "Other": 8}

STOP_WORDS = []


from dataclasses import dataclass
import os
from collections import defaultdict, namedtuple
from typing import Dict, List
import torch

from torch.utils.data import DataLoader, TensorDataset
from torch.nn.parallel import DistributedDataParallel

from tqdm import tqdm
import pickle as pkl
import numpy as np
from .utils import *
from botiverse.Theorizer.squad.squad_example import SquadExample, SquadProcessedExample
from transformers import (
    WEIGHTS_NAME,
    CONFIG_NAME,
    AdamW,
    GPT2Config,
    GPT2LMHeadModel,
    GPT2Tokenizer,
    get_linear_schedule_with_warmup,
)

SPECIAL_TOKENS = [
    "<sos>",
    "<eos>",
    "<paragraph>",
    "<clue>",
    "<answer>",
    "<style>",
    "<question>",
    "<pad>",
]
SPECIAL_TOKENS_DICT = {
    "bos_token": "<sos>",
    "eos_token": "<eos>",
    "pad_token": "<pad>",
    "unk_token": "<unk>",
    "additional_special_tokens": [
        "<paragraph>",
        "<clue>",
        "<answer>",
        "<style>",
        "<question>",
    ],
}
IGNORE_VALUE_OF_LM_HEADS = -100


@dataclass
class SquadGPT2Example:
    """
    A single example for the SQuAD dataset, processed for GPT-2 training.
    """

    input_ids: List[int]
    attention_mask: List[int] = None
    token_type_ids: List[int] = None
    lm_labels: List[int] = None


def prepare_squad_data_for_gpt2(tokenizer: GPT2Tokenizer, processed_examples: List[SquadProcessedExample]) -> List[SquadGPT2Example]:
    SOS, EOS, PARAGRAPH, CLUE, ANSWER, STYLE, QUESTION = tokenizer.convert_tokens_to_ids(
        SPECIAL_TOKENS[:-1])
    truncated_sequences = 0
    prepared_data = []

    for inst in tqdm(processed_examples):
        # Tokenize context, question, answer, question_type, and clue if available
        tokenized_context = tokenizer.encode(inst.context_text)
        tokenized_question = tokenizer.encode(inst.question_text)
        tokenized_answer = tokenizer.encode(inst.answer_text)
        tokenized_ans_prefix = tokenizer.encode(
            inst.context_text[: inst.answer_start + 1])
        tokenized_qtype = tokenizer.encode(inst.question_type)

        clue_exist = inst.clue_start is not None
        if clue_exist:
            tokenized_clue = tokenizer.encode(inst.clue_text)
            tokenized_clue_prefix = tokenizer.encode(
                inst.context_text[: inst.clue_start + 1])
        else:
            tokenized_clue = []

        # Calculate the total sequence length
        total_seq_len = (
            len(tokenized_context)
            + len(tokenized_answer)
            + len(tokenized_question)
            + len(tokenized_clue)
            + len(tokenized_qtype)
            + 6  # 6 special tokens, without pad or eos
        )

        # Truncate the sequence if it exceeds the GPT-2 model's input size
        if total_seq_len > tokenizer.max_model_input_sizes["gpt2"]:
            tokenized_context = tokenized_context[
                : -1 * (total_seq_len - tokenizer.max_model_input_sizes["gpt2"] + 1)
            ]
            truncated_sequences += 1

        # Calculate answer and clue positions in the tokenized context
        answer_position_tokenized = get_overlap_position(
            tokenized_context, tokenized_answer, tokenized_ans_prefix)
        if clue_exist:
            clue_position_tokenized = get_overlap_position(
                tokenized_context, tokenized_clue, tokenized_clue_prefix)
        else:
            clue_position_tokenized = (None, None)

        # Build input sequence and token_type_ids
        sequence = [SOS] + tokenized_context + [ANSWER] + tokenized_answer + [CLUE] + \
            tokenized_clue + [STYLE] + tokenized_qtype + \
            [QUESTION] + tokenized_question + [EOS]
        token_types = np.full(
            len(sequence), IGNORE_VALUE_OF_LM_HEADS, dtype=int)
        token_types[:len(tokenized_context) + 1] = PARAGRAPH
        token_types[answer_position_tokenized[0] +
                    1:answer_position_tokenized[1] + 1] = ANSWER
        if clue_exist:
            token_types[clue_position_tokenized[0] +
                        1:clue_position_tokenized[1] + 1] = CLUE
        token_types[len(tokenized_context) + len(tokenized_answer) + 2:len(
            tokenized_context) + len(tokenized_answer) + len(tokenized_clue) + 3] = CLUE
        token_types[len(tokenized_context) + len(tokenized_answer) + len(tokenized_clue) + 3:len(
            tokenized_context) + len(tokenized_answer) + len(tokenized_clue) + len(tokenized_qtype) + 4] = STYLE
        token_types[len(tokenized_context) + len(tokenized_answer) +
                    len(tokenized_clue) + len(tokenized_qtype) + 4:-1] = QUESTION

        # Build lm_labels
        lm_labels = np.full(len(sequence), IGNORE_VALUE_OF_LM_HEADS, dtype=int)
        lm_labels[-len(tokenized_question) - 1:-1] = tokenized_question
        lm_labels[-1] = EOS

        # Create a data point
        # TODO: Add attention_mask
        # I wrongly decided to use numpy arrays instead of lists for the data points, so I have to convert them back to lists here
        data_point = SquadGPT2Example(
            input_ids=list(sequence),
            token_type_ids=list(token_types),
            lm_labels=list(lm_labels),
        )
        # Add the data point to the prepared_data list
        prepared_data.append(data_point)

    return prepared_data


def prepare_and_pad_squad_data_for_gpt2(tokenizer: GPT2Tokenizer, processed_examples: List[SquadProcessedExample], max_len: int = None, padding: int = 0) -> List[SquadGPT2Example]:
    """
    Prepare and pad SQuAD data for GPT-2.

    This function tokenizes and processes the input data, builds the input sequence, token_type_ids, and lm_labels suitable for training GPT-2.
    It then pads the sequences to the maximum length in the dataset with the specified padding value.

    Args:
        tokenizer (GPT2Tokenizer): The GPT-2 tokenizer used to tokenize the text.
        processed_examples (List[SquadProcessedExample]): A list of SquadProcessedExample instances containing the processed SQuAD data.
        padding (int, optional): The padding value to use when padding the sequences. Default is 0.

    Returns:
        Dict[str, List[List[int]]]: A dictionary containing the prepared and padded data points with keys 'input_ids', 'token_type_ids', and 'lm_labels'.
    """
    prepared_data = prepare_squad_data_for_gpt2(tokenizer, processed_examples)

    max_l = max(len(x.input_ids) for x in prepared_data) if max_len is None else max_len

    data_point:SquadGPT2Example
    for data_point in prepared_data:
        data_point.input_ids = data_point.input_ids + [padding] * (max_l - len(data_point.input_ids))
        data_point.token_type_ids = data_point.token_type_ids + [padding] * (max_l - len(data_point.token_type_ids))
        data_point.lm_labels = data_point.lm_labels + [IGNORE_VALUE_OF_LM_HEADS] * (max_l - len(data_point.lm_labels))

    return prepared_data


def from_dict_to_squad_processed_example(data: Dict) -> SquadProcessedExample:
    """
    Convert a dictionary to a SquadProcessedExample instance.

    Args:
        data (Dict): A dictionary containing the data to convert.

    Returns:
        SquadProcessedExample: The SquadProcessedExample instance containing the data.
    """
    return SquadProcessedExample(
        context_text=data["paragraph"],
        question_text=data["question"],
        question_type=data["ques_type"],
        answer_text=data["answer"],
        answer_start=data["answer_start"],
        clue_text=data["clue"],
        clue_start=data["clue_start"],
        para_id=data["para_id"],
    )

def read_cached_processed_examples(filepath: str) -> List[SquadProcessedExample]:
    with open(filepath, "rb") as f:
        exmaples = pkl.load(f)
        exmaples = list(map(from_dict_to_squad_processed_example, exmaples))
    return exmaples

def test_getdataset():
    tokenizer = GPT2Tokenizer.from_pretrained("gpt2")
    tokenizer.add_tokens(SPECIAL_TOKENS)
    with open("squad/dataset/train.processed.pkl", "rb") as f:
        data = pkl.load(f)
        data = list(map(from_dict_to_squad_processed_example, data))
    data = prepare_and_pad_squad_data_for_gpt2(tokenizer, data, max_len=512)
    print(data[0])
    print(tokenizer.decode(data[0].input_ids))
    # paragraph = tokenizer.decode(data[0]["paragraph"])
    # print(data[0])
    # print(paragraph)

def test_tokenizer():
    a= [50257, 1026, 318, 257, 30069, 286, 262, 7128, 33955, 379, 406, 454, 8906, 11, 4881, 810, 262, 5283, 5335, 1128, 7241, 306, 4120, 284, 9281, 6206, 324, 5857, 311, 12944, 343, 516, 287, 1248, 3365, 13, 50261, 48615, 6206, 324, 5857, 311, 12944, 343, 516, 50260, 1169, 5283, 5335, 50262, 8727, 50263, 2514, 4150, 750, 262, 5283, 5335, 7910, 1656, 287, 1248, 3365, 287, 406, 454, 8906, 4881, 30, 50258]

    b= [50257, 1026, 318, 257, 30069, 286, 262, 7128, 33955, 379, 406, 454, 8906, 11, 4881, 810, 262, 5283, 5335, 1128, 7241, 306, 4120, 284, 9281, 6206, 324, 5857, 311, 12944, 343, 516, 287, 1248, 3365, 13, 50261, 48615, 6206, 324, 5857, 311, 12944, 343, 516, 50260, 1169, 5283, 5335, 50262, 8727, 50263, 2514, 4150, 750, 262, 5283, 5335, 7910, 1656, 287, 1248, 3365, 287, 406, 454, 8906, 4881, 30,50258]
    print(a==b)
    tokenizer=GPT2Tokenizer.from_pretrained("gpt2")
    tokenizer.add_tokens(SPECIAL_TOKENS)
    print(tokenizer.decode([8241]))
    print(tokenizer.decode([8727]))

if __name__ == "__main__":
    test_getdataset()

# TODO: remove this one after testing the above one for correctness, this one is for historical reasons
def prepare_squad_for_gpt2(tokenizer: GPT2Tokenizer, processed_examples: List[SquadProcessedExample], split) -> List[Dict]:

    truncated_sequences = 0
    inst: SquadProcessedExample
    for inst in tqdm(processed_examples):

        tokenized_context = tokenizer.encode(inst.context_text)
        tokenized_question = tokenizer.encode(inst.question_text)
        tokenized_answer = tokenizer.encode(inst.answer_text)
        tokenized_ans_prefix = tokenizer.encode(
            inst.context_text[: inst.answer_start + 1])
        tokenized_qtype = tokenizer.encode(inst.question_type)

        clue_exist = inst.clue_start is not None
        if clue_exist:
            tokenized_clue = tokenizer.encode(inst.clue_text)
            tokenized_clue_prefix = tokenizer.encode(
                inst.context_text[: inst.clue_start + 1])
        else:
            tokenized_clue = []

        total_seq_len = (
            len(tokenized_context)
            + len(tokenized_answer)
            + len(tokenized_question)
            + len(tokenized_clue)
            + len(tokenized_qtype)
            + 6  # 6 special tokens, without pad or eos
        )

        if total_seq_len > tokenizer.max_model_input_sizes["gpt2"]:
            # Heuristic to chop off extra tokens in paragraphs
            tokenized_context = tokenized_context[
                : -1 * (total_seq_len - tokenizer.max_model_input_sizes["gpt2"] + 1)
            ]
            truncated_sequences += 1
            assert (
                len(tokenized_context)
                + len(tokenized_answer)
                + len(tokenized_question)
                + len(tokenized_clue)
                + len(tokenized_qtype)
                + 6
                < tokenizer.max_model_input_sizes["gpt2"]
            )

        ans_prefix_ids = tokenizer.encode(tokenized_ans_prefix)
        answer_position_tokenized = get_overlap_position(
            tokenized_context, tokenized_answer, ans_prefix_ids
        )

        if clue_exist:
            clue_position_tokenized = get_overlap_position(
                tokenized_context, tokenized_clue, tokenized_clue_prefix
            )
        data = {
            "paragraph": tokenized_context,
            "question": tokenized_question,
            "answer": tokenized_answer,
            "answer_position_tokenized": answer_position_tokenized,
            "style": tokenized_qtype,
        }

    return data






"""
This module defines a GPT2 model with a custom head for language modeling. 
It is adapted from the original huggin face implementation. Some functions are taken as is, as they are necessary for the model to work, they override methods from the GPT2PreTrainedModel class, like the prepare_inputs_for_generation method, __reopen_input_ids, etc.

I implemented the forward function, with the same set of parameters as the original implementaion, they used during generation.
"""
from typing import *
import torch
import torch.nn as nn
from torch.nn import CrossEntropyLoss
from transformers import GPT2PreTrainedModel, GPT2Model
from transformers.modeling_outputs import CausalLMOutputWithCrossAttentions


class MyGPT2LMHeadModel(GPT2PreTrainedModel):
    _keys_to_ignore_on_load_missing = [
        r"linear1.weight", r"linear2.weight", r"linear3.weight", r"lm_head.weight"]
    _keys_to_ignore_on_load_unexpected = [
        r"h\.\d+\.attn\.masked_bias", r"h\.\d+\.attn\.bias"]

    def __init__(self, config):
        super().__init__(config)
        self.transformer = GPT2Model(config)
        self.linear1 = nn.Linear(config.n_embd, 1024, bias=False)
        self.linear2 = nn.Linear(1024, 512, bias=False)
        self.linear3 = nn.Linear(512, config.n_embd, bias=False)
        self.lm_head = nn.Linear(2048, config.vocab_size, bias=False)

        self.post_init()

    def get_output_embeddings(self):
        return self.lm_head

    def set_output_embeddings(self, new_embeddings):
        self.lm_head = new_embeddings

    def prepare_inputs_for_generation(self, input_ids, past_key_values=None, inputs_embeds=None, **kwargs):
        token_type_ids = kwargs.get("token_type_ids", None)
        # only last token for inputs_ids if past is defined in kwargs
        if past_key_values:
            input_ids = input_ids[:, -1].unsqueeze(-1)
            if token_type_ids is not None:
                token_type_ids = token_type_ids[:, -1].unsqueeze(-1)

        attention_mask = kwargs.get("attention_mask", None)
        position_ids = kwargs.get("position_ids", None)

        if attention_mask is not None and position_ids is None:
            # create position_ids on the fly for batch generation
            position_ids = attention_mask.long().cumsum(-1) - 1
            position_ids.masked_fill_(attention_mask == 0, 1)
            if past_key_values:
                position_ids = position_ids[:, -1].unsqueeze(-1)
        else:
            position_ids = None

        # if `inputs_embeds` are passed, we only want to use them in the 1st generation step
        if inputs_embeds is not None and past_key_values is None:
            model_inputs = {"inputs_embeds": inputs_embeds}
        else:
            model_inputs = {"input_ids": input_ids}

        model_inputs.update(
            {
                "past_key_values": past_key_values,
                "use_cache": kwargs.get("use_cache"),
                "position_ids": position_ids,
                "attention_mask": attention_mask,
                "token_type_ids": token_type_ids,
            }
        )
        return model_inputs

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[Tuple[Tuple[torch.Tensor]]] = None,
        attention_mask: Optional[torch.FloatTensor] = None,
        token_type_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, CausalLMOutputWithCrossAttentions]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Labels for language modeling. Note that the labels **are shifted** inside the model, i.e. you can set
            `labels = input_ids` Indices are selected in `[-100, 0, ..., config.vocab_size]` All labels set to `-100`
            are ignored (masked), the loss is only computed for labels in `[0, ..., config.vocab_size]`
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        transformer_outputs = self.transformer(
            input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_attention_mask,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        hidden_states = transformer_outputs[0]
        hidden_states = self.linear1(hidden_states)
        hidden_states = nn.functional.gelu(hidden_states)
        hidden_states = self.linear2(hidden_states)
        hidden_states = nn.functional.gelu(hidden_states)
        hidden_states = self.linear3(hidden_states)
        hidden_states = nn.functional.gelu(hidden_states)
        lm_logits = self.lm_head(hidden_states)

        loss = None
        if labels is not None:
            # move labels to correct device to enable model parallelism
            labels = labels.to(lm_logits.device)
            # Shift so that tokens < n predict n
            shift_logits = lm_logits[..., :-1, :].contiguous()
            shift_labels = labels[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = CrossEntropyLoss()
            loss = loss_fct(
                shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))

        if not return_dict:
            output = (lm_logits,) + transformer_outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return CausalLMOutputWithCrossAttentions(
            loss=loss,
            logits=lm_logits,
            past_key_values=transformer_outputs.past_key_values,
            hidden_states=transformer_outputs.hidden_states,
            attentions=transformer_outputs.attentions,
            cross_attentions=transformer_outputs.cross_attentions,
        )

    @staticmethod
    def _reorder_cache(
        past_key_values: Tuple[Tuple[torch.Tensor]], beam_idx: torch.Tensor
    ) -> Tuple[Tuple[torch.Tensor]]:
        """
        This function is used to re-order the `past_key_values` cache if [`~PreTrainedModel.beam_search`] or
        [`~PreTrainedModel.beam_sample`] is called. This is required to match `past_key_values` with the correct
        beam_idx at every generation step.
        """
        return tuple(
            tuple(past_state.index_select(0, beam_idx.to(past_state.device))
                  for past_state in layer_past)
            for layer_past in past_key_values
        )



def get_overlap_position(para_ids, ans_ids, ans_prefix_ids):
    """
    Get the position (start and end indices) of the overlapping region between a paragraph and an answer after the answer prefix.

    Args:
        para_ids (list): The paragraph token IDs.
        ans_ids (list): The answer token IDs.
        ans_prefix_ids (list): The prefix token IDs of the answer.

    Returns:
        tuple: A tuple representing the start and end indices of the overlapping region.
    """
    # Find the first index where the paragraph and answer prefix differ
    for i, (para_id, ans_prefix_id) in enumerate(zip(para_ids, ans_prefix_ids)):
        if para_id != ans_prefix_id:
            first_diff_index = i
            break
    else:
        first_diff_index = min(len(ans_prefix_ids), len(para_ids))

    # Calculate the end index of the overlapping region
    overlap_end_index = min(first_diff_index + len(ans_ids), len(para_ids))

    return (first_diff_index, overlap_end_index)


def pad_dataset(dataset, padding=0):
    """Pad the dataset. This could be optimized by defining a Dataset class and padd only batches but this is simpler."""
    max_l = max(len(x) for x in dataset["input_ids"])
    for name in MODEL_INPUTS:
        dataset[name] = [
            x + [padding if name != "lm_labels" else -100] * (max_l - len(x))
            for x in dataset[name]
        ]
    return dataset



from typing import List
from .dataloader import prepare_and_pad_squad_data_for_gpt2, read_cached_processed_data, SquadGPT2Example, SPECIAL_TOKENS
from .finetuned_model import MyGPT2LMHeadModel
import torch
from tqdm import tqdm
from transformers import (
    GPT2Config,
    GPT2TokenizerFast,
    get_linear_schedule_with_warmup,
)
from torch.utils.data import DataLoader, TensorDataset
# Training parameters and configuration
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
n_gpu = torch.cuda.device_count()
n_epochs = 10
lr = 5e-4
gradient_accumulation_steps = 4
max_norm = 1.0
train_batch_size = 4
valid_batch_size = 4
padding_length = 512
config = GPT2Config.from_pretrained("gpt2")
tokenizer = GPT2TokenizerFast.from_pretrained("gpt2")
tokenizer.add_tokens(SPECIAL_TOKENS)

model = MyGPT2LMHeadModel.from_pretrained("gpt2", config=config)
if not (next(model.parameters()).is_cuda):
    model.to(device)
model.resize_token_embeddings(len(tokenizer))
optimizer = torch.optim.Adam(model.parameters(), lr=lr)


def train(model, tokenizer, train_loader, val_loader, n_epochs, output_dir):
    """
    Train a GPT-2 model on the SQuAD dataset.
    
    Args:
        model: The GPT-2 model to be trained.
        tokenizer: The tokenizer used for tokenizing the data.
        train_loader: The DataLoader containing the training data.
        val_loader: The DataLoader containing the validation data.
        n_epochs: The number of epochs to train the model.
        output_dir: The directory where the model and tokenizer will be saved.
    """
    total_steps = len(train_loader) * n_epochs
    warmup_steps = int(0.1 * total_steps)  # 10% of total steps
    num_training_steps = len(train_loader) * \
        n_epochs // gradient_accumulation_steps

    optimizer = torch.optim.AdamW(model.parameters(), lr=lr)

    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=warmup_steps,
        num_training_steps=total_steps
    )

    global_step = 0
    for epoch in tqdm(range(n_epochs)):
        model.train()
        epoch_loss = 0
        for step, batch in tqdm(enumerate(train_loader)):
            optimizer.zero_grad()
            if n_gpu == 1:
                batch = tuple(t.to(device) for t in batch)
            cur_input_ids, cur_lm_labels, cur_token_type_ids = batch
            model_outputs = model(
                input_ids=cur_input_ids,
                labels=cur_lm_labels,
                token_type_ids=cur_token_type_ids,
            )
            loss = model_outputs[0]
            loss = loss / gradient_accumulation_steps
            epoch_loss += loss.item()

            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm)
            optimizer.step()
            scheduler.step()
            global_step += 1

        epoch_loss /= len(train_loader)

        # Evaluate on validation set
        model.eval()
        with torch.no_grad():
            val_loss = 0
            for batch in tqdm(val_loader):
                if n_gpu == 1:
                    batch = tuple(t.to(device) for t in batch)
                cur_input_ids, cur_lm_labels, cur_token_type_ids = batch
                model_outputs = model(
                    input_ids=cur_input_ids,
                    labels=cur_lm_labels,
                    token_type_ids=cur_token_type_ids,
                )

                lm_loss = model_outputs[0]
                loss = lm_loss / gradient_accumulation_steps
                val_loss += loss.item()
            val_loss /= len(val_loader)

        # Save final model and tokenizer
        epoch_output_dir = f"{output_dir}_gelu_{epoch}"
        model_to_save = model.module if hasattr(model, "module") else model
        model_to_save.save_pretrained(epoch_output_dir)
        tokenizer.save_pretrained(epoch_output_dir)


if __name__ == "__main__":
    output_dir = "/kaggle/working/model"
    train_examples = read_cached_processed_data(
        "squad/dataset/train.processed.pkl")
    ############################### Training Data ###############################
    train_features: List[SquadGPT2Example] = prepare_and_pad_squad_data_for_gpt2(
        processed_examples=train_examples,
        tokenizer=tokenizer,
        max_len=padding_length
    )
    all_input_ids = torch.tensor(
        [f.input_ids for f in train_features], dtype=torch.long)
    all_token_type_ids = torch.tensor(
        [f.token_type_ids for f in train_features], dtype=torch.long)
    all_lm_labels = torch.tensor(
        [f.lm_labels for f in train_features], dtype=torch.long)
    train_data = TensorDataset(
        all_input_ids, all_token_type_ids, all_lm_labels)
    train_dataloader = DataLoader(
        train_data, batch_size=train_batch_size)
    
    ############################### Validation Data ###############################
    valid_examples = read_cached_processed_data(
        "squad/dataset/dev.processed.pkl")
    valid_features: List[SquadGPT2Example] = prepare_and_pad_squad_data_for_gpt2(
        processed_examples=valid_examples,
        tokenizer=tokenizer,
        max_len=padding_length
    )
    all_input_ids = torch.tensor(
        [f.input_ids for f in valid_features], dtype=torch.long)
    all_token_type_ids = torch.tensor(
        [f.token_type_ids for f in valid_features], dtype=torch.long)
    all_lm_labels = torch.tensor(
        [f.lm_labels for f in valid_features], dtype=torch.long)
    valid_data = TensorDataset(
        all_input_ids, all_token_type_ids, all_lm_labels)
    valid_dataloader = DataLoader(
        valid_data, batch_size=valid_batch_size)

    train(model, tokenizer, train_dataloader, valid_dataloader, n_epochs, output_dir)



import torch
from botiverse.Theorizer.model.finetuned_model import MyGPT2LMHeadModel
from botiverse.Theorizer.model.dataloader import SPECIAL_TOKENS_DICT
from botiverse.Theorizer.squad.sample_data import select_with_default_sampel_probs
from transformers import GPT2Tokenizer
import os

current_file_dir = os.path.dirname(os.path.abspath(__file__))

def __prepare(context):
    sampled_infos = select_with_default_sampel_probs(context)
    instances = []
    for info in sampled_infos["selected_infos"]:
        for style in info["styles"]:
            for clue in info["clues"]:
                instances.append(
                    {
                        "paragraph": sampled_infos["context"],
                        "clue": clue.clue_text,
                        "answer": info["answer"]["answer_text"],
                        "style": style,
                    }
                )
    return instances


def generate(context, max_length=50):
    # Load the fine-tuned model and tokenizer
    
    model_path_or_name = os.path.join(current_file_dir,"model/pretrained-model")
    
    model = MyGPT2LMHeadModel.from_pretrained(
        model_path_or_name, ignore_mismatched_sizes=True
    )
    tokenizer = GPT2Tokenizer.from_pretrained(
        model_path_or_name, **SPECIAL_TOKENS_DICT
    )

    instances = __prepare(context)
    # print(instances)
    qa_dict = {}
    qa_dict["context"] = context
    qa_dict["qa"] = set()
    for inst in instances:
        paragraph = inst["paragraph"]
        clue = inst["clue"]
        answer = inst["answer"]
        style = inst["style"]
        input_sequence = (
            "<sos> "
            + paragraph
            + " <clue> "
            + clue
            + " <answer> "
            + answer
            + " <style> "
            + style
            + " <question> "
            + style
        )

        # Tokenize the input sequence
        input_ids = tokenizer.encode(input_sequence, return_tensors="pt")
        with torch.no_grad():
            # Generate the question
            generated = model.generate(
                input_ids=input_ids,
                max_length=max_length,
                eos_token_id=tokenizer.eos_token_id,
                pad_token_id=tokenizer.pad_token_id,
                num_beams=5,
                no_repeat_ngram_size=2,
                early_stopping=True,
                temperature=0.7,
            )
        question_start_index = (
            input_ids[0].tolist().index(
                tokenizer.convert_tokens_to_ids("<question>"))
        )

        # Slice the generated tensor to exclude the input sequence
        generated_question = generated[0, question_start_index + 1:]

        # Decode the generated question
        question = tokenizer.decode(
            generated_question, skip_special_tokens=True)

        qa_dict["qa"].add((question, answer))
    qa_dict["qa"] = list(qa_dict["qa"])
    return qa_dict


if __name__ == "__main__":
    context = "Bob is eating a delicious cake in Vancouver."
    qa_dict = generate(context)
    import json
    print(json.dumps(qa_dict,indent=4))



from typing import List, Set, Tuple, NewType
import numpy as np
import copy
from nltk.tree import Tree
import nltk
import spacy
import benepar
from dataclasses import dataclass
from .utils import *



@dataclass
class ClueInfo:
    clue_pos_tag: str
    clue_ner_tag: str
    clue_length: int
    clue_chunk: Tuple[str, str, List[str], int, int]
    clue_answer_dep_path_len: int
    padded_selected_clue_binary_ids: np.ndarray


QuestionType = NewType('QuestionType', str)


@dataclass
class SquadAugmentedExample:
    """
    A single training/test example for the Squad-Zhou dataset, after augmenting with clue info and question style.
    """
    context_text: str
    question_text: str
    answer_text: str
    question_type: QuestionType
    answer_pos_tag: str
    answer_ner_tag: str
    answer_length: int
    clue_info: ClueInfo


def __navigate(node: nltk.Tree):
    """
    Recursively navigates through an NLTK parse tree and extracts information about
    the tree depth, number of words, and chunks with their respective positions.

    Args:
        node (nltk.Tree): An NLTK parse tree node, generated from a constituency parser.

    Returns:
        tuple: A tuple containing the following elements:
            - max_depth (int): The maximum depth of the parse tree.
            - word_num (int): The total number of words in the parse tree.
            - chunklist (list): A list of tuples where each tuple contains the chunk's
                                POS tag and a list of its leaves with their respective
                                positions in the sentence.
    """

    """
    Disclaimer: This function is adapted from the original implementation with changes for readability, clarity and documentation
    """
    if type(node) is not Tree:
        return 1, 1, [("word", (node, 0))]

    # process leaf nodes
    for idx, _ in enumerate(node.leaves()):
        tree_location = node.leaf_treeposition(idx)
        non_terminal = node[tree_location[:-1]]
        if type(non_terminal[0]) is not str:
            non_terminal[0] = non_terminal[0] + [idx]
        else:
            non_terminal[0] = [non_terminal[0], idx]

    # process non-leaf nodes
    max_depth, word_num, chunklist = 0, 0, []
    for child in node:
        child_depth, child_num, child_chunklist = __navigate(child)
        max_depth = max(max_depth, child_depth + 1)
        word_num += child_num
        chunklist += child_chunklist

    # process current node
    cur_node_chunk = [(node.label(), node.leaves())]
    chunklist += cur_node_chunk
    return max_depth, word_num, chunklist


def chunks(sentence: str) -> List[Tuple[str, str, List[str], int, int]]:
    """
    Takes a sentence and returns a list of chunks with their respective NER tags,
    POS tags, words, and start and end positions in the sentence.

    Args:
        sentence (str): The input sentence to be parsed and chunked.

    Returns:
        list: A list of tuples where each tuple contains the following elements:
            - chunk_ner_tag (str): The NER tag of the chunk (e.g., 'PERSON', 'ORG').
            - chunk_pos_tag (str): The POS tag of the chunk (e.g., 'NP', 'VP').
            - leaves_without_position (list): A list of words in the chunk.
            - start (int): The start position of the chunk in the sentence.
            - end (int): The end position of the chunk in the sentence.
    """

    """
    Disclaimer: This function is adapted from the original implementation with changes for readability, clarity and documentation
    """
    tree = PARSER.parse(sentence)
    max_depth, node_num, orig_chunklist = __navigate(tree)
    spacy_document = NLP(sentence)
    chunklist = []
    for chunk in orig_chunklist:
        try:
            if chunk[0] == "word":
                continue
            chunk_pos_tag, leaves = chunk
            leaves_without_position = []
            position_list = []

            for v in leaves:
                if type(v) == list:  # Check if v is a list
                    wd = v[0]
                    idx = v[1:]
                    leaves_without_position.append(wd)
                    position_list.append(idx[0])
                else:
                    leaves_without_position.append(v)

            st = position_list[0]
            ed = position_list[-1]
            chunk_ner_tag = "UNK"
            chunk_text = " ".join(leaves_without_position)

            for ent in spacy_document.ents:
                if ent.text == chunk_text or chunk_text in ent.text:
                    chunk_ner_tag = ent.label_

            chunklist.append(
                (chunk_ner_tag, chunk_pos_tag, leaves_without_position, st, ed)
            )
        except:
            continue

    return chunklist


def __dfs(token_list, current_token, current_path, max_depth, related_tokens):
    if len(current_path) > max_depth:
        return
    if current_token in related_tokens and len(related_tokens[current_token]) <= len(current_path):
        return
    related_tokens[current_token] = current_path
    for token in token_list:
        if token.i != current_token:
            continue
        new_path = copy.deepcopy(current_path)
        try:
            new_path.append(token.dep_)
        except:
            continue
        __dfs(token_list, token.head.i, new_path, max_depth, related_tokens)
        for child in token.children:
            new_path = copy.deepcopy(current_path)
            new_path.append(child.head.dep_)
            __dfs(token_list, child.i, new_path, max_depth, related_tokens)


def get_dependency_paths(token_list: List[SpacyToken]):
    """
    Given a list of spaCy tokens, extract the dependency paths between different tokens.

    Args:
        doc (spacy.tokens.Doc): A spaCy document.

    Returns:
        dict: A dictionary mapping token indices to tokens.
        dict: A dictionary mapping token indices to related tokens and their dependency paths.
        list: A list of token texts.
    """
    index_to_token = {token.i: token for token in token_list}

    index_to_related = {}
    tokens = [token.text for token in token_list]

    for token in token_list:
        related_tokens = {}
        __dfs(
            token_list,
            token.i,
            [],
            len(token_list) - 1,
            related_tokens,
        )
        sorted_related = sorted(related_tokens.items(),
                                key=lambda x: len(x[1]))
        index_to_related[token.i] = sorted_related

    return index_to_token, index_to_related, tokens


def __tokenize_and_stem(sentence: str, condition_list: List[bool] = None) -> Tuple[List[str], List[str], List[SpacyToken]]:
    """
    Takes a sentence and returns a list of tokens, a list of stemmed tokens and a list of spacy tokens, only with the content words.

    Args:
        sentence (str): The input sentence to be tokenized and stemmed.

    Returns:
        tuple: A tuple of two lists where the first list contains the tokens and the second list contains the stemmed tokens.
    """

    if condition_list is not None:
        sentence_tokens = [
            token for i, token in enumerate(sentence) if condition_list[i]]
        spacy_tokens = None
        sentence_lemmas = None
    else:
        spacy_doc = NLP(sentence)
        spacy_tokens = [token for token in spacy_doc]
        
        sentence_tokens = [token.text for token in spacy_doc]
        sentence_lemmas = [token.lemma_ for token in spacy_doc]
    return sentence_tokens, sentence_lemmas, spacy_tokens


def __number_of_overlapping_tokens(sentence_tokens1: List[str], sentence_tokens2: List[str], condition_list: List[bool]) -> int:
    conditioned_sentence_1 = [tk for i, tk in enumerate(
        sentence_tokens1) if condition_list[i]]

    conditioned_sentence1_inetersect_sentence2 = [
        tk for tk in conditioned_sentence_1 if tk in sentence_tokens2]
    return len(conditioned_sentence1_inetersect_sentence2)


def extract_clue(sentence: str, question: str, answer: str, answer_start: int, config: InfoConfig = InfoConfig()) -> ClueInfo:
    """
    Given a sentence, question, answer, and the answer's starting position,
    this function extracts information about the clues related to the answer.

    Args:
        sentence (str): The sentence containing the answer.
        question (str): The question being asked.
        answer (str): The correct answer.
        answer_start (int): The starting position of the answer in the sentence.

    Returns:
        A Clue Info object holding all the clue information.
    """

    sentence = sentence.lower()
    question = question.lower()
    chunklist = chunks(sentence)

    sentence_tokens, sentence_lemmas, sentence_spacy_tokens = __tokenize_and_stem(
        sentence)
    question_tokens, question_lemmas, question_spacy_tokens = __tokenize_and_stem(
        question)
    idx2token, idx2related, context_tokens = \
        get_dependency_paths(sentence_spacy_tokens)

    
    # spans = find_token_spans_in_text(sentence, sentence_tokens)
    # answer_end = answer_start + len(answer)
    # answer_span = [idx for idx, span in enumerate(spans) if not (
    #     answer_end <= span[0] or answer_start >= span[1])]
    # y = answer_span[0]
    # answer_start = y

    clue_scores = []
    for chunk in chunklist:
        chunk_ner_tag, chunk_pos_tag, chunk_words, chunk_start, chunk_end = chunk

        chunk_content_words = [
            bool(word.lower() not in FUNCTION_WORDS) for word in chunk_words]
        chunk_text = " ".join(chunk_words).lower()

        chunk_tokens, _, _ = __tokenize_and_stem(
            chunk_words, chunk_content_words)
        chunk_lemmas = sentence_lemmas[chunk_start:chunk_end+1]
        chunk_lemmas = [lemma for i, lemma in enumerate(
            chunk_lemmas) if chunk_content_words[i]]
        no_tc = __number_of_overlapping_tokens(
            chunk_tokens, question_tokens, chunk_content_words)
        no_mc = __number_of_overlapping_tokens(
            chunk_lemmas, question_lemmas, chunk_content_words)
        binary_x = int(chunk_text in question)
        score = 0
        # chunk_lemmas_is_subset_from_question_lemmas = len(set(chunk_lemmas) & set(question_lemmas)) == len(set(chunk_lemmas))
        chunk_lemmas_is_subset_from_question_lemmas = \
            set(chunk_lemmas) <= set(question_lemmas) or \
            set(chunk_tokens) <= set(question_tokens)
        if chunk_lemmas_is_subset_from_question_lemmas and chunk_words[0].lower() not in NOT_BEGIN_TOKENS_FOR_ANSWER_CLUE and sum(map(lambda x: int(x), chunk_content_words)) > 0:
            score = no_tc*2+no_mc+binary_x
        clue_scores.append(score)

    # no clues were found
    padded_selected_clue_binary_ids = np.zeros([InfoConfig.sent_limit], dtype=np.float32)
    if not clue_scores or max(clue_scores) == 0:
        clue_chunk = None
        clue_pos_tag = "UNK"
        clue_ner_tag = "UNK"
        clue_length = 0
        clue_answer_dep_path_len = -1
    else:
        clue_chunk = clue_ner_tag, clue_pos_tag, chunk_words, clue_start, clue_end = \
            chunklist[clue_scores.index(max(clue_scores))]

        clue_answer_dep_path_len = abs(clue_start - answer_start)
        answer_related = idx2related[answer_start]
        clue_length = clue_end - clue_start + 1
        for tk_id, path in answer_related:
            if tk_id == clue_start:
                clue_answer_dep_path_len = len(path)

        if clue_start < InfoConfig.sent_limit and clue_end < InfoConfig.sent_limit:
                padded_selected_clue_binary_ids[clue_start: clue_end + 1] = 1

    return ClueInfo(
        clue_pos_tag=clue_pos_tag,
        clue_ner_tag=clue_ner_tag,
        clue_length=clue_length,
        clue_chunk=clue_chunk,
        clue_answer_dep_path_len=clue_answer_dep_path_len,
        padded_selected_clue_binary_ids=padded_selected_clue_binary_ids
    )


def extract_question_type_and_id(question, config: InfoConfig = InfoConfig()) -> Tuple[QuestionType, int]:
    """
    Given a question string, returns its type and associated id.

    Args:
        question (str): A question string.

    Returns:
        tuple: A tuple containing the question type (str) and its id (int).
    """

    # Split the question into words
    words = question.split()

    # Check if the question is an informational question
    for word in words:
        for i, question_type in enumerate(QUESTION_WORDS):
            if question_type.lower() in word.lower():
                return question_type, Q_TYPE2ID_DICT[question_type]

    # Check if the question is a boolean question
    if words[0].lower() in (q_type.lower() for q_type in BOOL_QUESTION_WORDS):
        return "boolean", Q_TYPE2ID_DICT["boolean"]

    # Return "other" if the question type is not found
    return "other", Q_TYPE2ID_DICT["other"]


def extract_clue_and_question_info(sentence: str, question: str, answer: str, answer_start: int, config: InfoConfig = InfoConfig) -> SquadAugmentedExample:
    """
    Extracts information about the question, answer, and clue from the provided sentence,
    question, and answer.

    Args:
        sentence (str): The sentence containing the answer.
        question (str): The question being asked.
        answer (str): The answer to the question.
        answer_start (int): The character index of the answer's start position in the sentence.
        config: A configuration object containing token limits and clue extraction settings.

    Returns:
        SquadAugmentedExample object containing extracted information about the question, answer, and clue.
    """

    # Process the input sentence and chunks
    sentence_spacydoc = NLP(sentence)
    chunklist = chunks(sentence)

    # Extract answer information
    answer_pos_tag, answer_ner_tag = "UNK", "UNK"
    for chunk in chunklist:
        if answer == " ".join(chunk[2]):
            answer_ner_tag = chunk[0]
            answer_pos_tag = chunk[1]
            break
    answer_length = len(answer.split())

    # Extract question type and id
    question_type, question_id = extract_question_type_and_id(question)

    # Get the answer start token index
    ans_sent_tokens = [token.text for token in sentence_spacydoc]
    spans = find_token_spans_in_text(sentence, ans_sent_tokens)
    answer_end = answer_start + len(answer)
    answer_span = []
    for idx, span in enumerate(spans):
        if not (answer_end <= span[0] or answer_start >= span[1]):
            answer_span.append(idx)
    answer_start_idx = answer_span[0]
    # Extract clue information
    clue_info = extract_clue(sentence, question, answer, answer_start_idx)

    example = SquadAugmentedExample(
        question_text=question,
        context_text=sentence,
        answer_text=answer,
        question_type=question_type,
        answer_pos_tag=answer_pos_tag,
        answer_ner_tag=answer_ner_tag,
        answer_length=answer_length,
        clue_info=clue_info,
    )

    return example


def test_chunks():
    # Test cases
    test_sentences = [
        "Beyoncé Giselle Knowles-Carter is an American singer, songwriter, and actress.",
        "Barack Obama was the 44th President of the United States.",
        "The Eiffel Tower is a wrought-iron lattice tower on the Champ de Mars in Paris, France."
    ]

    for sentence in test_sentences:
        print(chunks(sentence))


def test_extract_clue():
    # Test case 1
    test_sentence_1 = "Albert Einstein was a theoretical physicist who developed the theory of relativity."
    test_question_1 = "Who developed the theory of relativity?"
    test_answer_1 = "Albert Einstein"
    test_answer_start_1 = 0

    result_1 = extract_clue(test_sentence_1, test_question_1,
                            test_answer_1, test_answer_start_1)
    print(result_1)

    # Test case 2
    test_sentence_2 = "The capital of France is Paris, which is known for its culture, art, and landmarks."
    test_question_2 = "I apologize for the incomplete response. Here's the complete test case 2:"

    # Test case 2
    test_sentence_2 = "The capital of France is Paris, which is known for its culture, art, and landmarks."
    test_question_2 = "What is the capital of France?"
    test_answer_2 = "Paris"
    test_answer_start_2 = 21

    result_2 = extract_clue(test_sentence_2, test_question_2,
                            test_answer_2, test_answer_start_2)
    print(result_2)


def test_extract_question_type():
    # Test informational question types
    assert extract_question_type_and_id("Who are you?") == ("who", 1)
    assert extract_question_type_and_id("Where do you live?") == ("where", 3)
    assert extract_question_type_and_id("When were you born") == ("when", 4)
    assert extract_question_type_and_id("Why are you here?") == ("why", 5)
    assert extract_question_type_and_id(
        "Which one do you prefer?") == ("which", 6)
    assert extract_question_type_and_id("What is your name?") == ("what", 0)
    assert extract_question_type_and_id("How are you?") == ("how", 2)

    # Test boolean question types
    assert extract_question_type_and_id("Is it ok?") == ("boolean", 7)
    assert extract_question_type_and_id("Can you sleep?") == ("boolean", 7)
    assert extract_question_type_and_id("Should I leave?") == ("boolean", 7)
    assert extract_question_type_and_id(
        "Would you ..., please?") == ("boolean", 7)

    # Test "Other" question type
    assert extract_question_type_and_id(
        "Name the capital of France.") == ("other", 8)


def test_extract_clue_and_question_info():
    # Test case 1: A simple question and answer
    sentence1 = "The quick brown fox jumps over the lazy dog."
    question1 = "What color is the fox?"
    answer1 = "brown"
    answer_start1 = 10
    result1 = extract_clue_and_question_info(
        sentence1, question1, answer1, answer_start1)
    print(result1)
    assert result1.question_type == "what"
    assert result1.answer_pos_tag == "JJ"
    assert result1.answer_ner_tag == "UNK"
    assert result1.clue_info.clue_pos_tag == "ADJ"
    assert result1.clue_info.clue_ner_tag == "UNK"
    print("succeeded")


if __name__ == "__main__":
    # test_chunks()
    # test_extract_clue()
    # test_extract_question_type()
    test_extract_clue_and_question_info()



from dataclasses import dataclass
from typing import List, Tuple, TypeVar, Callable, Optional, Union, Type
import copy
import spacy
import benepar
import numpy as np
import math
import os

current_file_dir = os.path.dirname(os.path.abspath(__file__))
NLP = spacy.load('en_core_web_sm')
benepar.download('benepar_en3')
PARSER = benepar.Parser("benepar_en3")

FUNCTION_WORDS = set(
    [
        word.rstrip().lower()
        for word in open(
            os.path.join(current_file_dir,"function_words.txt"), "r", encoding="utf-8"
        ).readlines()
    ]
)
NOT_BEGIN_TOKENS_FOR_ANSWER_CLUE = set([
    "the",
    "of",
    "for",
    "to",
    "is",
    "are",
    "and",
    "was",
    "were",
    ",",
    "?",
    ";",
    "!",
    ".",
]
)

QUESTION_TYPES = [
    "who",
    "where",
    "when",
    "why",
    "which",
    "what",
    "how",
    "boolean",
    "other",
]
QUESTION_WORDS = ["who", "where", "when", "why", "which", "what", "how"]
BOOL_QUESTION_WORDS = [
    "am",
    "is",
    "was",
    "were",
    "are",
    "does",
    "do",
    "did",
    "have",
    "had",
    "has",
    "could",
    "can",
    "shall",
    "should",
    "will",
    "would",
    "may",
    "might",
]
Q_TYPE2ID_DICT = {
    "what": 0,
    "who": 1,
    "how": 2,
    "where": 3,
    "when": 4,
    "why": 5,
    "which": 6,
    "boolean": 7,
    "other": 8,
}



class SpacyDoc:
    """
    A stub class for the spacy.tokens.doc.Doc object, used for type hinting.
    """
    pass
class SpacyToken:
    """
    A stub class for the spacy.tokens object, used for type hinting.
    """
    pass


@dataclass
class InfoConfig:
    """
    Configuration for information extraction.
    """

    # style options
    num_sample_style: int = 2
    # answer options
    num_sample_answer: int = 5
    ans_len_bin_width: int = 3
    ans_len_min_val: int = 0
    ans_len_max_val: int = 30
    ans_limit: int = 30
    # clue options
    num_sample_clue: int = 2
    is_clue_topN: int = 20
    clue_dep_dist_bin_width: int = 2
    clue_dep_dist_min_val: int = 0
    clue_dep_dist_max_val: int = 20
    # question options
    ques_limit: int = 50
    # sampling options
    sent_limit: int = 100
    max_sample_times: int = 20


def find_token_spans_in_text(text: str, tokens: str) -> List[Tuple[int, int]]:
    """
    Get the character-level spans of the specified tokens in the input text.

    :param text: str, input text
    :param tokens: list of str, token texts to find in the input text
    :return: list of tuples, representing the character-level spans of each token in the input text
             Each tuple contains the start and end indices of the token in the text (end index exclusive)

    :raises Exception: If any of the specified tokens cannot be found in the input text
    """
    
    current = 0
    token_idx = 0
    spans = []
    text_len = len(text)
    token_len = len(tokens[token_idx])
    while current < text_len:
        if text[current:current + token_len] == tokens[token_idx]:
            spans.append((current, current + token_len))
            current += token_len
            token_idx += 1
            if token_idx >= len(tokens):
                break
            token_len = len(tokens[token_idx])
        else:
            current += 1

    if token_idx != len(tokens):
        print("Token {} cannot be found".format(tokens[token_idx]))
        raise Exception()

    return spans


def match_spans(pattern: str, input_text: str) -> List[Tuple[int, int]]:
    """
    Find all occurrences of the given pattern in the input text and return the character-level spans.

    :param pattern: str, the pattern to match in the input text
    :param input_text: str, the input text where the pattern will be searched
    :return: list of tuples, each tuple represents the character-level span of the pattern in the input text
             Each tuple contains the start and end indices of the pattern in the text (end index exclusive)
    """
    pattern_len = len(pattern)
    text_len = len(input_text)
    spans = []

    for idx in range(text_len - pattern_len + 1):
        if input_text[idx:idx + pattern_len] == pattern:
            spans.append((idx, idx + pattern_len))

    return spans


def normalize(text: str) -> str:
    """
    Normalize the given text by replacing `` with " and '' with ".
    """
    return text.replace("''", '" ').replace("``", '" ')

def token_to_char_indices(sentence):
    """
    Generate character index ranges for each token in a given sentence using spaCy.

    Args:
        sentence (str): The input sentence to be tokenized.

    Returns:
        token_to_char_range (dict): A dictionary where keys are token indices and values are tuples representing the start
                                    and end indices (inclusive) of the token in the sentence.
        char_to_token (dict): A dictionary where keys are character indices and values are the corresponding token indices
                              in the sentence.
    """
    doc = NLP(sentence)
    token_to_char_range = {}
    char_to_token = {}
    for token in doc:
        start_idx, end_idx = token.idx, token.idx + len(token.text) - 1
        token_to_char_range[token.i] = (start_idx, end_idx)
        
        for char_idx in range(start_idx, end_idx + 1):
            char_to_token[char_idx] = token.i

    return token_to_char_range, char_to_token


def weighted_sample(choices, probs):
    """
    Sample from `choices` with probability according to `probs`.

    Args:
        choices (list): A list of elements to sample from.
        probs (list): A list of probabilities corresponding to each element in `choices`. 
                      The probabilities don't need to be normalized.

    Returns:
        any: A randomly sampled element from `choices` based on the provided probabilities.
    """
    probs = np.array(probs) / sum(probs)
    cumulative_probs = np.cumsum(probs)
    r = np.random.random()

    # Find the index of the first cumulative probability greater than or equal to r
    index = np.searchsorted(cumulative_probs, r)
    return choices[index]

def value_to_bin(input_val: int, min_val: int, max_val: int, bin_width: int):
    """
    Determine the bin index for the given input value, based on a binned range between min_val and max_val with a specified bin width.

    Args:
        input_val (float): The input value to be binned.
        min_val (float): The minimum value of the range.
        max_val (float): The maximum value of the range.
        bin_width (float): The width of each bin in the range.

    Returns:
        int: The bin index of the input value.
            - If the input value is within the range, the function returns the corresponding bin index.
            - If the input value is greater than the maximum value, the function returns the index of the last bin + 1.
            - If the input value is less than the minimum value, the function returns -1.
    """
    """
    Disclaimer this function is adapted from the original implementation
    """
    if min_val <= input_val <= max_val:
        return math.ceil((input_val - min_val) / bin_width)
    elif input_val > max_val:
        return math.ceil((max_val - min_val) / bin_width) + 1
    else:
        return -1

def str_find(text, tklist):
    """
    Searches for a sequence of tokens in a given text string, allowing for spaces between characters.

    The function takes a text string and a list of tokens as input, and returns the start and end character
    indices of the sequence of tokens in the text. If the sequence is not found within the text, the function
    returns (-1, -1). Spaces between characters in the input text are allowed and do not affect the search.

    Args:
        text (str): The input text string in which to search for the sequence of tokens.
        tklist (list): A list of tokens representing the sequence to search for in the text. Each token is a string.

    Returns:
        tuple: A tuple containing two integer values:
               - The start character index of the sequence in the text, or -1 if the sequence is not found.
               - The end character index of the sequence in the text, or -1 if the sequence is not found.

    Example:
        >>> text = "This is an example text."
        >>> tklist = ["ex", "am", "ple"]
        >>> str_find(text, tklist)
        (11, 18)
    """
    search_str = "".join(tklist)
    tk1 = tklist[0]
    pos = text.find(tk1)

    while pos >= 0 and pos < len(text):
        i, j = pos, 0
        while i < len(text) and j < len(search_str):
            if text[i] == " ":
                i += 1
                continue
            if text[i] == search_str[j]:
                i += 1
                j += 1
            else:
                break

        if j == len(search_str):
            return pos, i - 1

        newpos = text[pos + 1:].find(tk1)
        pos = pos + 1 + newpos if newpos >= 0 else -1

    return -1, -1

def test_find_token_spans_in_text():
    text = "Welcome hello world"
    tokens = ["hello", "world"]
    spans = find_token_spans_in_text(text, tokens)
    assert spans == [(8, 13), (14, 19)]


def test_match_spans():
    pattern = "hello"
    input_text = "Welcome hello world"
    spans = match_spans(pattern, input_text)
    assert spans == [(8, 13)]


if __name__ == "__main__":
    print(match_spans("hello", "Welcome hello world"))



from typing import List, Tuple, Dict, Any
from collections import Counter
from dataclasses import dataclass
import os

import numpy as np
from .utils import *
from .info_extractor import *
import pickle as pkl

current_file_dir = os.path.dirname(os.path.abspath(__file__))

@dataclass
class AnswerSample:
    answer_text: str
    char_st: int
    char_ed: int
    st: int
    ed: int
    answer_bio_ids: List[str]
    answer_pos_tag: str
    answer_ner_tag: str


@dataclass
class ClueSample:
    clue_text: str
    clue_binary_ids: np.ndarray


def select_answers(chunklist, sentence, sample_probs, config=InfoConfig()) -> List[AnswerSample]:
    """
    Select multiple answer chunks from a given list of chunks based on their probability.

    Args:
        chunklist (list): A list of chunks, where each chunk is a tuple containing NER tag, POS tag,
                          token leaves, start index, and end index.
        sentence (str): The input sentence from which the chunks are extracted.
        sample_probs (dict): A dictionary containing the probabilities of different answer conditions.
        config (InfoConfig, optional): A configuration object containing parameters for the sampling process.

    Returns:
        list: A list of sampled answers, where each answer is a tuple containing answer text, character start index,
              character end index, token start index, token end index, answer BIO tags, POS tag, and NER tag.
    """
    token2idx, idx2token = token_to_char_indices(sentence)
    # I can write haskell, can you?
    # chunk === chunk_ner_tag, chunk_pos_tag, leaves_without_position, st, ed
    chunk: Tuple[str, str, List[str], int, int]
    a_probs = [
        sample_probs["a"]
        ["_".join(["-".join([chunk[0], chunk[1]]),
                   str(value_to_bin(
                       abs(chunk[3] - chunk[4] + 1),
                       config.ans_len_min_val,
                       config.ans_len_max_val,
                       config.ans_len_bin_width
                   ))])]
        if chunk[2][0].lower() not in NOT_BEGIN_TOKENS_FOR_ANSWER_CLUE
        else 1
        for chunk in chunklist
    ]

    sampled_chunk_ids = set(
        weighted_sample(list(range(len(chunklist))), a_probs)
        for _ in range(config.max_sample_times)
    )
    sentence_spacydoc = NLP(sentence)

    def process_chunk(chunk):
        chunk_ner_tag, chunk_pos_tag, leaves, st, ed = chunk
        context = sentence
        char_st, char_ed = str_find(context, leaves)
        if char_st < 0:
            return None
        answer_text = context[char_st: char_ed + 1]
        st, ed = idx2token[char_st], idx2token[char_ed]
        answer_bio_ids = ["O"] * len(sentence_spacydoc)
        answer_bio_ids[st: ed + 1] = ["I"] * (ed - st + 1)
        answer_bio_ids[st] = "B"
        char_st, char_ed = token2idx[st][0], token2idx[ed][1]
        return AnswerSample(
            answer_text=answer_text,
            char_st=char_st,
            char_ed=char_ed,
            st=st,
            ed=ed,
            answer_bio_ids=answer_bio_ids,
            answer_pos_tag=chunk_pos_tag,
            answer_ner_tag=chunk_ner_tag,
        )

    sampled_answers = [
        answer for answer in (process_chunk(chunklist[chunk_id]) for chunk_id in sampled_chunk_ids) if answer
    ]
    return sampled_answers[:config.num_sample_answer]


def select_questions(ans: AnswerSample, sample_probs, config=InfoConfig()):
    """
    Select question styles based on the answer's POS and NER tags, given sample probabilities.

    Args:
        ans (AnswerSample): A tuple containing information about the answer, including its text, indices, BIO tags,
                            POS tag, and NER tag.
        sample_probs (dict): A dictionary containing the probabilities of different question styles.
        config (InfoConfig, optional): A configuration object containing the maximum number of sampling attempts and
                                       the desired number of question styles to sample.

    Returns:
        list: A list of sampled question styles.
    """
    a_tag = "-".join([ans.answer_pos_tag, ans.answer_ner_tag])

    # Get style probabilities
    s_probs = [
        sample_probs["s|c,a"].get("_".join([s, a_tag]), 1) for s in QUESTION_TYPES
    ]

    # Sample question styles
    sampled_styles = []
    for _ in range(config.max_sample_times):
        sampled_s = weighted_sample(QUESTION_TYPES, s_probs)
        if sampled_s not in sampled_styles:
            sampled_styles.append(sampled_s)
            if len(sampled_styles) >= config.num_sample_style:
                break

    return sampled_styles


def select_clues(chunklist, doc: SpacyDoc, ans: AnswerSample, sample_probs, config=InfoConfig()):
    """
    Select clues from a list of chunks based on the dependency distance and the probability of the chunk given the answer.

    Args:
        chunklist (list): A list of chunks, each containing NER tag, POS tag, text, start index, and end index.
        doc (spacy.tokens.Doc): A SpaCy document containing the tokens of the sentence.
        ans (AnswerSample): A tuple containing information about the answer, including its text, indices, BIO tags,
                            POS tag, and NER tag.
        config (InfoConfig, optional): A configuration object containing the maximum number of sampling attempts and
                                       the desired number of clues to sample.

    Returns:
        list: A list of sampled clues, with each clue containing its text and binary ids.
    """
    st, idx2related = ans.st, get_dependency_paths(doc)[1]
    context_tokens = [token.text for token in doc]

    # Calculate chunk probabilities
    c_probs = []
    for chunk in chunklist:
        c_tag = "-".join([chunk[1], chunk[0]])
        dep_dist = min(abs(chunk[3] - st), min(len(path)
                       for tk_id, path in idx2related[st] if tk_id == chunk[3]))
        dep_dist_bin = value_to_bin(dep_dist, config.clue_dep_dist_min_val,
                                    config.clue_dep_dist_max_val, config.clue_dep_dist_bin_width)

        if chunk[2][0].lower() not in NOT_BEGIN_TOKENS_FOR_ANSWER_CLUE:
            c_probs.append(sample_probs["c|a"].get(
                "_".join([c_tag, str(dep_dist_bin)]), 1))
        else:
            c_probs.append(1)

    # Sample clues
    chunk_ids = list(range(len(chunklist)))
    sampled_clue_chunk_ids = []
    for _ in range(config.max_sample_times):
        sampled_chunk_id = weighted_sample(chunk_ids, c_probs)
        if sampled_chunk_id not in sampled_clue_chunk_ids:
            sampled_clue_chunk_ids.append(sampled_chunk_id)
            if len(sampled_clue_chunk_ids) >= config.num_sample_clue:
                break

    # Extract clue details
    sampled_clues = []
    for chunk_id in sampled_clue_chunk_ids:
        chunk = chunklist[chunk_id]
        clue_start, clue_end = chunk[3], chunk[4]
        clue_text = " ".join(context_tokens[clue_start: clue_end + 1])
        clue_binary_ids = [0] * len(doc)
        clue_binary_ids[clue_start: clue_end + 1] = [1] * \
            (clue_end - clue_start + 1)
        sampled_clues.append(ClueSample(clue_text=clue_text,
                             clue_binary_ids=clue_binary_ids))

    return sampled_clues


def select(sentence, sample_probs, config=InfoConfig()):
    sampled_infos = []
    chunklist = chunks(sentence)
    doc = NLP(sentence)
    for ans in select_answers(chunklist, sentence, sample_probs):
        info = {
            "answer": {
                "answer_text": ans.answer_text,
                "char_start": ans.char_st,
                "char_end": ans.char_ed,
                "answer_bio_ids": ans.answer_bio_ids,
                "answer_chunk_tag": ans.answer_pos_tag,
            },
            "styles": None,
            "clues": None,
        }
        # sample styleselect
        styles = select_questions(ans, sample_probs)
        info["styles"] = list(styles)

        # sample clue
        selected_clues = select_clues(chunklist, doc, ans, sample_probs)
        info["clues"] = selected_clues

        sampled_infos.append(info)

    result = {
        "context": sentence,
        "selected_infos": sampled_infos,
        "ans_sent_doc": doc.text,
    }
    return result


def read_sample_probs(sample_probs_path):
    with open(sample_probs_path, "rb") as f:
        sample_probs = pkl.load(f)
    return sample_probs

def select_with_default_sampel_probs(sentence):
    sample_probs_path = os.path.join(current_file_dir,"sample_probs.pkl")
    sample_probs = read_sample_probs(sample_probs_path)
    selection = select(
        "Bob is eating a delicious cake in Vancouver.", sample_probs)
    return selection

def test():
    sample_probs_path = "botiverse/Theorizer/squad/sample_probs.pkl"
    sample_probs = read_sample_probs(sample_probs_path)
    selection = select(
        "Bob is eating a delicious cake in Vancouver.", sample_probs)
    print(selection)


if __name__ == "__main__":
    test()



import os
# from botiverse.Theorizer import generate
import gdown
import zipfile

curr_dir = os.path.dirname(os.path.abspath(__file__))

model_dir = os.path.join(curr_dir,"model")
squad_dir = os.path.join(curr_dir,"squad")

if not os.path.exists(os.path.join(squad_dir,"sample_probs.pkl")):
    print("Sample probs not found. Downloading Theorizer sample probs...")
    url = "https://drive.google.com/uc?id=1UjZaqM9jf9nzeK1R7WdSSLNEVKdxJOHO"
    gdown.download(url,os.path.join(squad_dir,"sample_probs.pkl"), quiet=False)
    print("Done.")

if not os.path.exists(os.path.join(model_dir,"pretrained-model")):
    print("Weights not found. Downloading Theorizer weights...")
    model_path = os.path.join(model_dir,"pretrained-model")
    url = "https://drive.google.com/drive/folders/1rUvMP1HdE_H4TAMG8HxHT6z5Y0ZOnBsg"
    gdown.download_folder(url,output=model_path,quiet=False)
    print("Done.")

from spacy.cli import download
download('en_core_web_sm')

import nltk
nltk.download('punkt')